@@ -1,4 +1,4 @@
-# Copyright 2004-2019 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2020 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -19,17 +19,18 @@
 # OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 # WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 
-from __future__ import print_function
+from __future__ import division, absolute_import, with_statement, print_function, unicode_literals
+from renpy.compat import *
 
 # This file contains displayables that move, zoom, rotate, or otherwise
 # transform displayables. (As well as displayables that support them.)
 import math
-import types  # @UnresolvedImport
+import types # @UnresolvedImport
 
-import renpy.display  # @UnusedImport
+import renpy.display # @UnusedImport
 from renpy.display.layout import Container
 
-import renpy.display.accelerator
+from renpy.display.accelerator import transform_render
 
 # The null object that's used if we don't have a defined child.
 null = None
@@ -110,7 +111,14 @@ class TransformState(renpy.object.Object):
     xtile = 1
     ytile = 1
     last_angle = None
+    xsize = None
+    ysize = None
+    fit = None
     maxsize = None
+    matrixcolor = None
+    shader = None
+    mesh = False
+    blur = None
 
     def __init__(self):
         self.alpha = 1
@@ -146,9 +154,17 @@ class TransformState(renpy.object.Object):
         self.crop_relative = False
         self.corner1 = None
         self.corner2 = None
-        self.size = None
+
+        self.xsize = None
+        self.ysize = None
+        self.fit = None
         self.maxsize = None
 
+        self.matrixcolor = None
+        self.shader = None
+        self.mesh = False
+        self.blur = None
+
         self.delay = 0
 
         self.debug = None
@@ -170,6 +186,12 @@ class TransformState(renpy.object.Object):
         self.inherited_xanchor = None
         self.inherited_yanchor = None
 
+        for i in uniforms:
+            setattr(self, i, None)
+
+        for i in gl_properties:
+            setattr(self, i, None)
+
     def take_state(self, ts):
 
         self.nearest = ts.nearest
@@ -191,7 +213,10 @@ class TransformState(renpy.object.Object):
         self.crop_relative = ts.crop_relative
         self.corner1 = ts.corner1
         self.corner2 = ts.corner2
-        self.size = ts.size
+
+        self.xsize = ts.xsize
+        self.ysize = ts.ysize
+        self.fit = ts.fit
         self.maxsize = ts.maxsize
 
         self.xpan = ts.xpan
@@ -199,11 +224,22 @@ class TransformState(renpy.object.Object):
         self.xtile = ts.xtile
         self.ytile = ts.ytile
 
+        self.matrixcolor = ts.matrixcolor
+        self.shader = ts.shader
+        self.mesh = ts.mesh
+        self.blur = ts.blur
+
         self.last_angle = ts.last_angle
 
         self.debug = ts.debug
         self.events = ts.events
 
+        for i in uniforms:
+            setattr(self, i, getattr(ts, i))
+
+        for i in gl_properties:
+            setattr(self, i, getattr(ts, i))
+
         # Take the computed position properties, not the
         # raw ones.
         (self.inherited_xpos,
@@ -263,7 +299,10 @@ class TransformState(renpy.object.Object):
         diff2("crop_relative", newts.crop_relative, self.crop_relative)
         diff2("corner1", newts.corner1, self.corner1)
         diff2("corner2", newts.corner2, self.corner2)
-        diff2("size", newts.size, self.size)
+
+        diff2("xsize", newts.xsize, self.xsize)
+        diff2("ysize", newts.ysize, self.ysize)
+        diff2("fit", newts.fit, self.fit)
         diff2("maxsize", newts.maxsize, self.maxsize)
 
         diff4("xpos", newts.xpos, newts.inherited_xpos, self.xpos, self.inherited_xpos)
@@ -280,9 +319,20 @@ class TransformState(renpy.object.Object):
         diff2("xtile", newts.xtile, self.xtile)
         diff2("ytile", newts.ytile, self.ytile)
 
+        diff2("matrixcolor", newts.matrixcolor, self.matrixcolor)
+
+        # It doesn't make sense to interpolate these.
+        # diff2("shader", newts.shader, self.shader)
+        # diff2("mesh", newts.mesh, self.mesh)
+
+        diff2("blur", newts.blur, self.blur)
+
         diff2("debug", newts.debug, self.debug)
         diff2("events", newts.events, self.events)
 
+        for i in uniforms:
+            diff2(i, getattr(newts, i), getattr(self, i))
+
         return rv
 
     def get_placement(self, cxoffset=0, cyoffset=0):
@@ -398,6 +448,16 @@ class TransformState(renpy.object.Object):
 
     offset = property(get_offset, set_offset)
 
+    def get_size(self):
+        return self.xsize, self.ysize
+
+    def set_size(self, value):
+        if value is None:
+            value = (None, None)
+        self.xsize, self.ysize = value
+
+    size = property(get_size, set_size)
+
     def set_xcenter(self, value):
         self.xpos = value
         self.xanchor = 0.5
@@ -476,7 +536,11 @@ class Transform(Container):
     crop_relative = Proxy("crop_relative")
     corner1 = Proxy("corner1")
     corner2 = Proxy("corner2")
+
+    xsize = Proxy("xsize")
+    ysize = Proxy("ysize")
     size = Proxy("size")
+    fit = Proxy("fit")
     maxsize = Proxy("maxsize")
 
     delay = Proxy("delay")
@@ -495,6 +559,11 @@ class Transform(Container):
     xtile = Proxy("xtile")
     ytile = Proxy("ytile")
 
+    matrixcolor = Proxy("matrixcolor")
+    shader = Proxy("shader")
+    mesh = Proxy("mesh")
+    blur = Proxy("blur")
+
     debug = Proxy("debug")
     events = Proxy("events")
 
@@ -582,7 +651,7 @@ class Transform(Container):
             self.arguments = { }
 
             # Fill self.arguments with a
-            for k, v in kwargs.iteritems():
+            for k, v in kwargs.items():
 
                 prefix = ""
                 prop = k
@@ -608,7 +677,7 @@ class Transform(Container):
                         prefix = new_prefix
 
             if "" in self.arguments:
-                for k, v in self.arguments[""].iteritems():
+                for k, v in self.arguments[""].items():
                     setattr(self.state, k, v)
 
         else:
@@ -667,7 +736,7 @@ class Transform(Container):
             if d is None:
                 continue
 
-            for k, v in d.iteritems():
+            for k, v in d.items():
                 setattr(state, k, v)
 
         return None
@@ -823,6 +892,7 @@ class Transform(Container):
 
         self.child = child
         self.children = [ child ]
+
         self.child_st_base = self.st
 
         child.per_interact()
@@ -853,6 +923,8 @@ class Transform(Container):
         self.active = True
 
     # The render method is now defined in accelerator.pyx.
+    def render(self, width, height, st, at):
+        return transform_render(self, width, height, st, at)
 
     def event(self, ev, x, y, st):
 
@@ -868,7 +940,7 @@ class Transform(Container):
         if not offsets:
             return None
 
-        for i in xrange(len(self.children)-1, -1, -1):
+        for i in range(len(self.children) - 1, -1, -1):
 
             d = children[i]
             xo, yo = offsets[i]
@@ -993,9 +1065,6 @@ class Transform(Container):
         rv.take_execution_state(self)
         rv._unique()
 
-        if rv.child and rv.child._duplicatable:
-            rv.child = rv.child._duplicate(None)
-
         return rv
 
     def _in_current_store(self):
@@ -1017,9 +1086,6 @@ class Transform(Container):
         self.update_state()
 
 
-Transform.render = types.MethodType(renpy.display.accelerator.transform_render, None, Transform)
-
-
 class ATLTransform(renpy.atl.ATLTransformBase, Transform):
 
     def __init__(self, atl, child=None, context={}, parameters=None, **properties):
@@ -1050,3 +1116,47 @@ class ATLTransform(renpy.atl.ATLTransformBase, Transform):
     def _show(self):
         super(ATLTransform, self)._show()
         self.execute(self, self.st, self.at)
+
+
+uniforms = set()
+gl_properties = set()
+
+
+def add_uniform(name):
+    """
+    Adds a uniform with `name` to Transform and ATL.
+    """
+
+    if name in uniforms:
+        return
+
+    if not name.startswith("u_"):
+        return
+
+    if name.startswith("u_renpy"):
+        return
+
+    if name in renpy.gl2.gl2draw.standard_uniforms:
+        return
+
+    uniforms.add(name)
+    setattr(TransformState, name, None)
+    setattr(Transform, name, Proxy(name))
+    renpy.atl.PROPERTIES[name] = renpy.atl.any_object
+
+
+def add_gl_property(name):
+    """
+    Adds a GL property with `name` to Transform and ATL.
+    """
+
+    if name in gl_properties:
+        return
+
+    gl_properties.add(name)
+    setattr(TransformState, name, None)
+    setattr(Transform, name, Proxy(name))
+    renpy.atl.PROPERTIES[name] = renpy.atl.any_object
+
+
+add_gl_property("gl_color_mask")
